#ifndef	FSM_H
#define	FSM_H
// $Id: fsm.h,v 1.8 2001/08/08 12:33:32 dvermeir Exp $

#include <iostream>
#include <string>
#include <set>
#include <map>
#include <stdexcept>

#include "ref.h" // reference-counted pointers


//! Class representing finite state machines.
/*! A more generic approach would use the types State and Symbol
    as template parameters, i.e.
    \code
    template<typename State, typename Symbol>
    class Fsm {
    ...
    }
    \endcode
*/
class Fsm {
public:
  //! Type of state.
  /*! This is a simple solution. However, it is probably sufficient.
      After all: the required operations on a State are few:
      we shoul be able to print its name, build containers of
      State's (for this we need e.g. State::operator<), check
      whether states are equal (State::operator==).
      All of these (and more) are supported by string.
  */
  typedef string State;
  //! Type of set of states.
  /*! E.g. the Fsm::delta function will return a reference to an
      object of this type.
  */
  typedef set<State> STATES;
  //! Type of symbol.
  /*! This is a simple solution. However, it is probably sufficient.
      After all: the required operations on a Symbol are few:
      we shoul be able to print it, build containers of
      Symbol's (for this we need e.g. Symbol::operator<), check
      whether symbols are equal (Symbol::operator==).
      All of these (and more) are supported by string.
  */
  typedef string Symbol;
  //! Type of set of symbols.
  typedef set<Symbol> SYMBOLS;
  //! Type used to store transitions.
  /*! 
      E.g. if delta(p,a) contains q, then
      \code
      TRANSITIONS	ts;

      ts[make_pair(p,a)].count(q) != 0
      \endcode

      The make_pair can be avoided by using Fsm::delta.

      A more efficient representation could be devised for
      deterministic automata.
  */
  typedef map<pair<State,Symbol>,STATES > TRANSITIONS;

  //! Return special ``epsilon'' symbol.
  /*! Technically, <code>epsilon()</code> is
      not a symbol but a string.
  */
  static const Symbol& epsilon();

  //! Static factory method replacing constructor (which is protected).
  /*! We return a ref<Fsm> because this is convenient to use
      for parser & optimizer classes and it saves the bother
      of managing real pointers.
  */
  static ref<Fsm> create(const State& initial_state);
  //! Destructor, probably a noop.
  virtual ~Fsm();

  //! Add state.
  /*! This function is automatically called e.g. by add_transition.
      Note that adding the same state twice has no effect.
  */
  Fsm& add_state(const State& s);
  //! Add symbol.
  /*! This function is automatically called e.g. by add_transition.
      Note that adding the same symbol twice has no effect.
  */
  Fsm& add_symbol(const Symbol& a);
  //! Add final state, also does <code>add_state(s)</code>.
  /*!  Note that adding the same state twice has no effect.
  */
  Fsm& add_final_state(const State& s);
  //! Add transition.
  /*! Also calls <code>add_state(q1)</code>,
      <code>add_state(q2)</code>, and
      <code>add_symbol(a)</code>.

      Note that adding the same transition twice has no effect.
  */
  Fsm& add_transition(const State& q1,const Symbol& a, const State& q2);

  //! Return initial state.
  const State& initial_state() const { return initial_state_; }
  //! Return final states.
  const STATES& final_states() const { return final_states_; }
  //! Return all states.
  const STATES& states() const { return states_; }

  //! Is this a final state?
  bool is_final(const State& s) const;

  //! Delta function.
  /*! \code
      delta(s,a)
      \endcode
      returns set of states that the machine may be in after receiving
      input <code>a</code> from state <code>s</code>.

      \warning The return reference is probably not valid after
      a subsequent update of the transitions_ data member.
  */
  const STATES&	    delta(const State&, const Symbol&) const;
  //! Deterministic delta function.
  /*! Only useful if the machine is deterministic.
     
      The function will throw a runtime_error for undefined transitions
      or non-deterministic transitions.
  */
  const State&	    ddelta(const State&, const Symbol&) const
  							throw(runtime_error);

  //! Is this a deterministic machine?
  /*! Returns true iff there are no epsilon transitions and
      the size of delta(q,a) is exactly 1 for each known transition.
  */
  bool is_deterministic() const;

  //! Return iterator pointing to first transition.
  TRANSITIONS::const_iterator	begin() const { return transitions_.begin(); }
  //! Return iterator pointing just after last transition.
  TRANSITIONS::const_iterator	end() const { return transitions_.end(); }

  //! Return iterator pointing to first state.
  STATES::const_iterator    sbegin() const { return states_.begin(); }
  //! Return iterator pointing just after last state.
  STATES::const_iterator    send() const { return states_.end(); }

  //! Return iterator pointing to first final state.
  STATES::const_iterator    fbegin() const { return final_states_.begin(); }
  //! Return iterator pointing just after last final state.
  STATES::const_iterator    fend() const { return final_states_.end(); }

  //! Return iterator pointing to first symbol in alphabet.
  SYMBOLS::const_iterator    abegin() const { return alphabet_.begin(); }
  //! Return iterator pointing just after last symbol in alphabet.
  SYMBOLS::const_iterator    aend() const { return alphabet_.end(); }

  //! Print textual representation of this fsm.
  friend ostream& operator<<(ostream& os, const Fsm& fsm);

protected:
  //! Constructor, parameter is initial state.
  /*! Requiring the initial state as a parameter for the constructor
      saves a lot of checking. Now we can be sure that every
      Fsm object has an initial state. 

      Note that the other member data of Fsm, namely the final states
      and the transitions, are sets which could be empty without
      complicating the code.
  */
  Fsm(const State& initial_state);

private:
  //! Initial state.
  State		initial_state_;
  //! Final states, use set for reasonably fast access.
  STATES	final_states_;
  //! All states.
  STATES	states_;
  //! All symbols, i.e. alphabet of fsm.
  SYMBOLS	alphabet_;

  //! Transitions.
  /*! E.g. if q' in delta(q,a) then 
      \code
      transitions_[make_pair(q,a)].count(q') = 1.
      \endcode
  */
  TRANSITIONS	transitions_;

};

//! Auxiliary function: print set of states.
ostream&
operator<<(ostream&, const Fsm::STATES& Q);
#endif
