// $Id: main.C,v 1.11 2001/08/18 10:20:34 dvermeir Exp $
#include <fstream>

#include "fsm.h"
#include "parser.h"
#include "optimizer.h"
#include "mkdfa.h"

/*! \mainpage

    De volgende klassen worden gedefinieerd:

    <ul>
    <li> Fsm representeert een automaat.  Er zijn memberfuncties 
         om de (eind-, begin- of andere) staten op te halen, om
	 het alfabet te kennen etc. Tevens zijn er functies om
	 de automaat op te bouwen of te wijzigen: Fsm::add_state etc.

	 De overgangsfunctie is
	 beschikbaar in een normale (Fsm::delta) en een
	 deterministische (Fsm::ddelta) versie. Deze laatste
	 kan enkel gebruikt worden voor deterministische automaten.

         In een meer gesophisticeerd ontwerp zou men intern een
	 meer efficiente implementatie kunnen gebruiken voor
	 de overgangen van een deterministische automaat.

    <li> Parser is een hulpklasse die in staat is om een 
         automaat op te bouwen uit een beschrijving op een
	 input stream. 
	 Enkel de publieke interface van Fsm wordt
	 gebruikt, wat de Parser onafhankelijk maakt van de
	 implementatie van Fsm.

	 Parser gebruikt nocommentstream, een stream met
	 een filterstreambuf die commentaar wegfiltert,
	 zie <a
	 href="http://tinf2.vub.ac.be/cpp/chapter-10/index.html">
	 hoofstuk 10</a> van het
	 <a href="http://tinf2.vub.ac.be/cpp/index.html>boek</a>.

    <li> DfaMaker is een hulpklasse die een deterministische
         Fsm kan bouwen die equivalent is met een gegeven
	 nondeterministische automaat. 
	 Enkel de publieke interface van Fsm wordt
	 gebruikt, wat DfaMaker onafhankelijk maakt van de
	 implementatie van Fsm.

    <li> Optimizer is een hulpklasse die een optimale
         Fsm kan bouwen die equivalent is met een gegeven
	 deterministische automaat. 
	 Enkel de publieke interface van Fsm wordt
	 gebruikt, wat Optimizer onafhankelijk maakt van de
	 implementatie van Fsm.
     </ul>

   Er zijn twee hoofdprogramma's:
   <ul>
   <li> fsm (broncode in main.C)

	De functionaliteit van dit programma wordt beschreven
	in de 
<a
href="http://tinf2.vub.ac.be/~dvermeir/courses/structuur2/stru2-2000b.html">opgave</a>.

	De volgende figuur (gegenereerd door fsm2dot en 
	<a
	href="http://www.research.att.com/sw/tools/graphviz/">dot</a>)
	laat de input automaat zien voor
	\code
	fsm -o nfa0.txt onfa0.txt
	\endcode

	\image html nfa0.gif "Input voor fsm -o nfa0.txt onfa0.txt"
	\image latex nfa0.ps "Input voor fsm -o nfa0.txt onfa0.txt"

	Het resultaat is te zien in de volgende figuur.
	\image html onfa0.gif "Output van fsm -o nfa0.txt onfa0.txt"
	\image latex onfa0.ps "Output van fsm -o nfa0.txt onfa0.txt"

   <li> fsm2dot (broncode in fsm2dot.C)

   	Dit programma is een filter die een automaat
	(in het formaat herkend door Parser) beschrijving
	vertaalt naar een graaf specificatie voor
	gebruikt door het programma
	<a href="http://www.research.att.com/sw/tools/graphviz/">
	dot</a> dat dan weer kan gebruikt worden om een
	grafische voorstelling van de automaat te genereren.
   </ul>
*/

//! Main function for the fsm program.
int
main(int argc, char* argv[]) {
static const string USAGE("fsm [-o] inputfile outputfile [inputstring trace]");

try {
  // There must be at least two arguments.
  
  if (argc<3) {
    cerr << USAGE << endl;
    return 1;
    }
  
  // Check whether optimization is requested.
  
  bool optimize = ( string(argv[1]) == "-o" );
  
  // If optimization is requested, at least 3 arguments must be present.
  if (optimize && (argc<4)) {
    cerr << USAGE << endl;
    return 1;
    }
  
  int fn_index(1); // index of first argument after optional '-o'
  
  // If optimization is requested, the first filename argumengt is at index 2.
  
  if (optimize)
    ++fn_index;
  
  // Is tracing requested? 
  
  bool tracing(false);
  
  if (argc > fn_index+2) { // There are arguments beyond input/output automaton.
    tracing = true;
    if (argc != fn_index + 4) { // need exactly 2 filenames
      cerr << USAGE << endl;
      return 1;
      }
    }
  
  // Determine input and output filenames.
  
  const char* input_fnm(argv[fn_index]);
  const char* output_fnm(argv[fn_index+1]);
  
  // Try to open the input file.
  
  ifstream  ifs(input_fnm);
  
  if (!ifs) {
    cerr << "Cannot open \"" << input_fnm << "\" for reading" << endl;
    return 2;
    }
  
  // Try to open the output file.
  
  ofstream  ofs(output_fnm);
  if (!ofs) {
    cerr << "Cannot open \"" << output_fnm << "\" for writing" << endl;
    return 2;
    }
  
  
  // The input automaton is *fsm.
  ref<Fsm>	fsm = Parser(ifs).parse();
  // The deterministic version of *fsm will be called *dfsm.
  ref<Fsm>	dfsm;
  // The optimized version of *dfsm will be called *ofsm.
  ref<Fsm>	ofsm;
  
  // Check whether parsing went ok.
  
  if (!fsm) {
    cerr << "Syntax error in \"" << input_fnm << "\"" << endl;
    return 3;
    }
  
  // Generate dfsm out of fsm.
  
  if (fsm->is_deterministic()) // already deterministic.
    dfsm = fsm;
  else
    dfsm = DfaMaker(*fsm).mkdfa();
  
  // Optimize dfsm, yielding ofsm, if requested.
  
  if (optimize) 
    ofsm = Optimizer(*dfsm).optimize();
  else
    ofsm = dfsm;
  
  ofs << *ofsm;
  
  // Run trace, if needed.
  
  if (tracing) {
    const char* trace_in_fn(argv[fn_index+2]);
    const char* trace_out_fn(argv[fn_index+3]);
  
    ifstream	tifs(trace_in_fn);
  
    if (!tifs) {
      cerr << "Cannot open \"" << trace_in_fn << "\" for reading" << endl;
      return 2;
      }
  
    ofstream	tofs(trace_out_fn);
  
    if (!tofs) {
      cerr << "Cannot open \"" << trace_out_fn << "\" for writing" << endl;
      return 2;
      }
  
    try { // There could be an exception, e.g. because of a bad Symbol.
      Fsm::State	q(ofsm->initial_state());
      Fsm::Symbol	a;
  
      while (tifs >> a) {
        tofs << q << ": " << a << " => ";
        q = ofsm->ddelta(q,a);
        tofs << q << "\n";
        }
  
      if (ofsm->is_final(q))
        tofs << "accepted";
      else
        tofs << "not accepted";
      tofs << "\n";
      }
    catch (exception& e) {
      tofs << "error: " << e.what() << "\n";
      cerr << "error: " << e.what() << endl;
      return 3;
      }
    }
  
  return 0;
  }
catch (exception& e) {
  cerr << "Error: " << e.what() << endl;
  return -1;
  }
}
