// $Id: optimizer.C,v 1.5 2001/08/08 12:33:33 dvermeir Exp $

#include "fsm.h"
#include "optimizer.h"


Optimizer::Optimizer(const Fsm& fsm, ostream* debug): fsm_(fsm), debug_(debug) {
}

bool&
Optimizer::distinct(const Fsm::State& q1, const Fsm::State& q2) {
// The distinct_ matrix is symmetric, we only store distinct_[q1,q2]
// if q1 > q2.
if (q1 < q2)
  return distinct_[make_pair(q2,q1)];
else
  return distinct_[make_pair(q1,q2)];
}

ref<Fsm>
Optimizer::optimize() {
// Optimization algorithm.

if (!fsm_.is_deterministic())
  return ref<Fsm>();

// Initialization of distinct_: any final state is distinct from any
// non-final state.

for (Fsm::STATES::const_iterator i= fsm_.sbegin(); i != fsm_.send(); ++i)
  for (Fsm::STATES::const_iterator j = fsm_.sbegin(); j!=i; ++j)
    distinct(*i,*j) = ( fsm_.is_final(*i) != fsm_.is_final(*j) );

// Fixpoint computation of distinct_: i is distinct from j if
// delta(i,a) is distinct from delta(j,a) for some a.

bool change(true);

while (change) {
  change = false;
  for (Fsm::STATES::const_iterator i = fsm_.sbegin(); i != fsm_.send(); ++i)
    for (Fsm::STATES::const_iterator j = fsm_.sbegin(); j!=i; ++j)
      if (distinct(*i,*j))
        ;
      else { 
        // Check if distinct(delta(*i,a),delta(*j,a)) for some a.
	for (Fsm::SYMBOLS::const_iterator a = fsm_.abegin(); a!= fsm_.aend(); ++a) {
	  Fsm::State	 qi(fsm_.ddelta(*i,*a));
	  Fsm::State	 qj(fsm_.ddelta(*j,*a));
	  if (distinct(qi,qj)) {
	    distinct(*i,*j) = change = true;
	    break; // out of for loop through alphabet
	    }
	  }
	}
        
  }

if (debug_) {
  for (Fsm::STATES::const_iterator i = fsm_.sbegin(); i != fsm_.send(); ++i)
    for (Fsm::STATES::const_iterator j = fsm_.sbegin(); j!=i; ++j) 
      *debug_ << "distinct(" << *i << "," << *j << ") = "
              << ( distinct(*i,*j) ? "true" : "false" ) << "\n";
  }

// Create new dfa: it suffices to select a single representative q for
// each class Q of undistinguishable states and define the transitions
// new_delta(q,a) = p where p is a representative of the class of
// delta(q,a).

ref<Fsm> ofsm(Fsm::create(eq_class(fsm_.initial_state())));

// Define final states: just add a representative of the class of each
// origvinal final state. Duplicates will automatically be refused by
// Fsm::add_final_state functions.
for (Fsm::STATES::const_iterator i = fsm_.fbegin(); i != fsm_.fend(); ++i) 
  ofsm->add_final_state(eq_class(*i));

// Define transitions: just translate delta(q,a) = p to delta(q',a) =
// p' where q' is the representative of the class of q and p' is the
// representative of the class of p.

for (Fsm::STATES::const_iterator i = fsm_.sbegin(); i != fsm_.send(); ++i) 
  for (Fsm::SYMBOLS::const_iterator a = fsm_.abegin(); a!= fsm_.aend(); ++a) 
    ofsm->add_transition(eq_class(*i), *a, eq_class(fsm_.ddelta(*i,*a)));

return ofsm;
}

Fsm::State
Optimizer::eq_class(const Fsm::State& q) {
// Compute the equivalence class of q and return its representative,
// i.e. the minimal one in the order defined on states.
Fsm::STATES	e;
e.insert(q);

for (Fsm::STATES::const_iterator i = fsm_.sbegin(); i != fsm_.send(); ++i)
  if ( (*i!=q) && (!distinct(*i,q)) )
    e.insert(*i);

if (debug_)
  *debug_ << "eq_class(" << q << ") = " << e << "\n";

return *e.begin();
}
